package encapsulation;

import jst.parsetree.JstParseTreeBranch;
import jst.symtab.Block;
import jst.symtab.ClassType;
import jst.symtab.ExecutableCode;
import jst.symtab.FieldDecl;
import jst.symtab.FieldInitializer;
import jst.symtab.OperationDecl;
import jst.symtab.ParameterizedClassType;
import jst.symtab.ParameterizedInterfaceType;
import jst.symtab.TypedDecl;
import jst.symtab.UserType;
import jst.symtab.Decl;
import jst.symtab.link.IdLookup;
import jst.symtab.link.OperationLookup;
import jst.symtab.link.TypeLink;
import jst.symtab.visitor.*;
import java.lang.reflect.Modifier;
import java.util.*;

import parser.parsetree.SyntaxTree;
import parser.parsetree.SyntaxTreeLeaf;

import sun.awt.image.ImageWatched.Link;

public class EncapsulationAnalysisVisitor extends CompositionVisitor {
	
	private String expressionName = "Expression";
	private String expressionNameName = "ExpressionName";
	private String expressionStatementName = "ExpressionStatement";
	private String lhsName = "LeftHandSide";

	private int[] fieldCounts;	//total number of variables by access modifier
	
	private int[] fieldAccesses;	//total number of variable accesses by access modifier
	
	private int totalFields;	//total number of variables
	
	private enum AccessModifier {PRIVATE, PACKAGE, PROTECTED, PUBLIC};
	
	private Map<UserType, List<Integer>> classFieldCounts;	//number of variables in each class
	
	private Map<UserType, Map<FieldDecl, Integer>> fieldAccessCounts;	//number of accesses for each variable, ordered by class
	
	private int totalFieldAccesses;	//total number of variable accesses
	private int fieldAccessesOtherClass;	//total number of variable accesses from outside the class
	private int fieldAccessesOtherObject;	//total number of variable accesses from outside the object
	
	private int publicFieldAccesses;		//accesses to public variables from outside of the class
	private int publicFieldAccessesOtherObject; //accesses to public variables from outside the object
	private int packageFieldAccessesOther;		//accesses to package variables from outside of the class
	private int packageFieldAccessesOtherObject; //accesses to package variables from outside the object
	private int protectedFieldAccessesAncestor; //accesses to protected variables in ancestors
	private int protectedFieldAccessesAncestorObject; //accesses to protected variables in ancestors from the same object
	private int protectedFieldAccessesOther; //accesses to protected variables that are not in ancestors
	private int protectedFieldAccessOtherObject; 	//accesses to protected variables from outside the object
	private int privateFieldAccesses; 	//access to private variables of another object in the same class

	
	private int[] methodCounts;	//total number of methods by access modifier
	
	private int[] methodAccesses;	//total number of method accesses by access modifier
	
	private int totalMethods;	//total number of methods
	
	private Map<UserType, List<Integer>> methodClassCounts;	//number of methods in each class
	
	private Map<UserType, Map<OperationDecl, Integer>> methodAccessCounts;	//number of accesses for each method, ordered by class
	
	private int totalMethodAccesses;	//total number of method accesses	
	private int methodAccessesOtherClass;	//total number of method accesses from outside the class
	private int methodAccessesOtherObject;	//total number of method accesses from outside the object
	
	private int publicMethodAccesses;		//accesses to public methods from outside of the class
	private int publicMethodAccessesOtherObject; //accesses to public methods from outside the object
	private int packageMethodAccessesOtherObject;	//accesses to package methods in ancestors
	private int packageMethodAccessesOther;	//accesses to package methods that are not in ancestors
	private int protectedMethodAccessesAncestor;	//accesses to protected methods in ancestors
	private int protectedMethodAccessesAncestorObject; //accesses to protected methods in ancestors from the same object
	private int protectedMethodAccessesOther;	//accesses to protected methods that are not in ancestors
	private int privateMethodAccesses;	//access to private methods of another object in the same class
	
	private int protectedMethodAccessOtherObject; 	//accesses to protected methods from outside the object
	//record the accesses and where they came from
	private Map<FieldDecl, Set<UserType>> fieldAccessMap;
	private Map<OperationDecl, Set<UserType>> methodAccessMap;
	
	//record the fields which are visible for a particular class
	//and the classes a field is visible from
	private Map<FieldDecl, Set<ClassType>> fieldVisibleFrom;
	private Map<ClassType, Set<FieldDecl>> fieldCanSee;
	
	//record the methods which are visible for a particular class
	//and the classes a method is visible from
	private Map<OperationDecl, Set<ClassType>> methodVisibleFrom;
	private Map<ClassType, Set<OperationDecl>> methodCanSee;
	
	public EncapsulationAnalysisVisitor() {
		fieldAccessMap = new HashMap<FieldDecl, Set<UserType>>();
		methodAccessMap = new HashMap<OperationDecl, Set<UserType>>();
		fieldVisibleFrom = new HashMap<FieldDecl, Set<ClassType>>();
		fieldCanSee = new HashMap<ClassType, Set<FieldDecl>>();
		
		fieldCounts = new int[AccessModifier.values().length];
		fieldAccesses = new int[AccessModifier.values().length];
		classFieldCounts = new HashMap<UserType, List<Integer>>();
		fieldAccessCounts = new HashMap<UserType, Map<FieldDecl, Integer>>();
		totalFields = 0;
		totalFieldAccesses = 0;
		publicFieldAccesses = 0;
		privateFieldAccesses = 0;
		protectedFieldAccessesAncestor = 0;
		protectedFieldAccessesOther = 0;
		protectedFieldAccessesAncestorObject = 0;
		packageFieldAccessesOtherObject = 0;
		packageFieldAccessesOther = 0;
		publicFieldAccessesOtherObject = 0;
		protectedFieldAccessOtherObject = 0;
		fieldAccessesOtherClass = 0;
		fieldAccessesOtherObject = 0;
		
		methodCounts = new int[AccessModifier.values().length];
		methodAccesses = new int[AccessModifier.values().length];
		methodVisibleFrom = new HashMap<OperationDecl, Set<ClassType>>();
		methodCanSee = new HashMap<ClassType, Set<OperationDecl>>();
		
		totalMethodAccesses = 0;
		totalMethods = 0;
		methodClassCounts = new HashMap<UserType, List<Integer>>();
		methodAccessCounts = new HashMap<UserType, Map<OperationDecl, Integer>>();
		publicMethodAccesses = 0;
		publicMethodAccessesOtherObject = 0;
		protectedMethodAccessesAncestor = 0;
		protectedMethodAccessesOther = 0;
		packageMethodAccessesOtherObject = 0;
		packageMethodAccessesOther = 0;
		privateMethodAccesses = 0;
		protectedMethodAccessOtherObject = 0;
		protectedMethodAccessesAncestorObject = 0;
		methodAccessesOtherClass = 0;
		methodAccessesOtherObject = 0;
	}
	
	public void visitClassType(ClassType classType) {
		if(classType instanceof ParameterizedClassType){
			return;
		}
		
		super.visitClassType(classType);
        
        if(!fieldCanSee.containsKey(classType)){
        	fieldCanSee.put(classType, new HashSet<FieldDecl>());
        }
        if(!methodCanSee.containsKey(classType)){
        	methodCanSee.put(classType, new HashSet<OperationDecl>());
        }
	}
	
	public void visitOperationDecl(OperationDecl operationDecl) {
		if(operationDecl.getOwnerUserType() instanceof ParameterizedClassType ||
				operationDecl.getOwnerUserType() instanceof ParameterizedInterfaceType){
			return;
		}
		
		super.visitOperationDecl(operationDecl);
        
        if(operationDecl.getOwnerUserType() == null || operationDecl.getOwnerUserType().getSource() == null) {
        	return;
        }
        
        if(!methodVisibleFrom.containsKey(operationDecl)){
        	methodVisibleFrom.put(operationDecl, new HashSet<ClassType>());
        }
        
        //store information about the method
        if(!methodAccessMap.containsKey(operationDecl)){
        	methodAccessMap.put(operationDecl, new HashSet<UserType>());
        }
        
        totalMethods++;
        
        int modifier = operationDecl.getModifier();
        int accessModifier = getAccessModifier(modifier);
        
        methodCounts[accessModifier]++;
        
        UserType classType = operationDecl.getOwnerUserType();
        if(methodClassCounts.containsKey(classType)){
        	List<Integer> list = methodClassCounts.get(classType);
        	list.set(accessModifier, list.get(accessModifier) + 1);
        }
        else {
        	List<Integer> list = new ArrayList<Integer>();
        	for(int i = 0; i < 4; i++) {
        		list.add(0);
        	}
        	list.set(accessModifier, 1);
        	methodClassCounts.put(classType, list);
        }
        
        //more in depth info about variable access
        //this code is needed here to make sure that we also find variables
        //with no accesses
		if(methodAccessCounts.containsKey(classType)){
			Map<OperationDecl, Integer> map = methodAccessCounts.get(classType);
			if(!map.containsKey(operationDecl)){
				map.put(operationDecl, 0);
			}
		}
		else {
			Map<OperationDecl, Integer> map = new HashMap<OperationDecl, Integer>();
			map.put(operationDecl, 0);
			methodAccessCounts.put(classType, map);
		}
    }
	
	public void visitFieldDecl(FieldDecl fieldDecl) {
		if(fieldDecl.getOwnerUserType() instanceof ParameterizedClassType ||
				fieldDecl.getOwnerUserType() instanceof ParameterizedInterfaceType){
			return;
		}
		
		super.visitFieldDecl(fieldDecl);
        
        //ignore library classes
        if(fieldDecl == null || fieldDecl.getOwnerUserType().getSource() == null){
        	return;
        }
        
        if(!fieldVisibleFrom.containsKey(fieldDecl)){
        	fieldVisibleFrom.put(fieldDecl, new HashSet<ClassType>());
        }
        
        //store information about the field
        if(!fieldAccessMap.containsKey(fieldDecl)){
        	fieldAccessMap.put(fieldDecl, new HashSet<UserType>());
        }
        
        //increment variable count
        totalFields++;
        
        int modifier = fieldDecl.getModifier();
        int accessModifier = getAccessModifier(modifier);
        
        fieldCounts[accessModifier]++;
        
        UserType classType = fieldDecl.getOwnerUserType();
        if(classFieldCounts.containsKey(classType)){
        	List<Integer> list = classFieldCounts.get(classType);
        	list.set(accessModifier, list.get(accessModifier) + 1);
        }
        else {
        	List<Integer> list = new ArrayList<Integer>();
        	for(int i = 0; i < 4; i++) {
        		list.add(0);
        	}
        	list.set(accessModifier, 1);
        	classFieldCounts.put(classType, list);
        }
        
        //more in depth info about variable access
        //this code is needed here to make sure that we also find variables
        //with no accesses
		if(fieldAccessCounts.containsKey(classType)){
			Map<FieldDecl, Integer> map = fieldAccessCounts.get(classType);
			if(!map.containsKey(fieldDecl)){
				map.put(fieldDecl, 0);
			}
		}
		else {
			Map<FieldDecl, Integer> map = new HashMap<FieldDecl, Integer>();
			map.put(fieldDecl, 0);
			fieldAccessCounts.put(classType, map);
		}
    }
	
	private int getAccessModifier(int modifier) {
		int accessModifier;
		if(Modifier.isPrivate(modifier)){
        	accessModifier = AccessModifier.PRIVATE.ordinal();
        }
        else if(Modifier.isProtected(modifier)){
        	accessModifier = AccessModifier.PROTECTED.ordinal();
        }
        else if(Modifier.isPublic(modifier)){
        	accessModifier = AccessModifier.PUBLIC.ordinal();
        }
        else {
        	accessModifier = AccessModifier.PACKAGE.ordinal();
        }
		return accessModifier;
	}
	
    public void visitExecutableCode(ExecutableCode executableCode) {
    	if(executableCode.getOwnerUserType() instanceof ParameterizedClassType ||
    			executableCode.getOwnerUserType() instanceof ParameterizedInterfaceType){
			return;
		}
    	
    	super.visitExecutableCode(executableCode);
    	
    	if(executableCode.getOwnerUserType() == null || executableCode.getSource() == null) {
    		return;
    	}
    	
    	Set<jst.symtab.link.Link> refersTo = executableCode.getReferences();
    	Iterator<jst.symtab.link.Link> refs = refersTo.iterator();
    	
		while (refs.hasNext()) {
			jst.symtab.link.Link link = refs.next();
			
			do {			
				visitExecutableCodeDecl(link, executableCode);
				
				if(link instanceof OperationLookup) {
					OperationLookup lookup = (OperationLookup) link;
					
					for(TypeLink argLink : lookup.getArgLinks()){
						argLink.resolve();
						visitExecutableCodeDecl(argLink, executableCode);
					}
				}
				
				//go to the inner link
				link = getNextLink(link);
			} while(link != null);
		}
	}
    
    private void visitExecutableCodeDecl(jst.symtab.link.Link link, ExecutableCode executableCode) {
    	//get the decl the link refers to
		Decl decl = link.getDecl();
		
		if(decl == null) {
			return;
		}
		if(decl.getOwnerUserType() == null || decl.getSource() == null){
			return;
		}
		
    	//get the expression used to refer to the decl
		String expression =  getExpression(link, decl);
		if(expression == null) {
			return;
		}

		//owner of decls is different for generic classes
		UserType owner = decl.getOwnerUserType();
		if(decl.getOwnerUserType() instanceof ParameterizedClassType ||
				decl.getOwnerUserType() instanceof ParameterizedInterfaceType){
			owner = (UserType)decl.getOwnerUserType().getOwnerDecl();
		}
		
		//field access
		if(decl instanceof FieldDecl) {					
			FieldDecl field = (FieldDecl) decl;
			
			//store information about the field access
	        if(!fieldAccessMap.containsKey(field)){
	        	Set<UserType> set = new HashSet<UserType>();
	        	set.add(executableCode.getOwnerUserType());
	        	fieldAccessMap.put(field, set);
	        }
	        else {
	        	Set<UserType> set = fieldAccessMap.get(field);
	        	set.add(executableCode.getOwnerUserType());
	        	fieldAccessMap.put(field, set);
	        }
			
			totalFieldAccesses++;
			int modifier = getAccessModifier(field.getModifier());
			fieldAccesses[modifier]++;
			
			//calculate the total number of accesses from outside the object / class
			if(!owner.equals(executableCode.getOwnerUserType())){
				fieldAccessesOtherClass++;
			}
			if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
					&& !owner.equals(executableCode.getOwnerUserType())){
				fieldAccessesOtherObject++;
			}
			else {
				int accesses = getOtherObjectAccesses(expression);
				fieldAccessesOtherObject += accesses;
			}
			
			//more in depth info about variable access
			if(fieldAccessCounts.containsKey(owner)){
				Map<FieldDecl, Integer> map = fieldAccessCounts.get(owner);
				if(map.containsKey(field)){
					map.put(field, map.get(field) + 1);
				}
				else {
					map.put(field, 1);
				}
				fieldAccessCounts.put(owner, map);
			}
			else {
				Map<FieldDecl, Integer> map = new HashMap<FieldDecl, Integer>();
				map.put(field, 1);
				fieldAccessCounts.put(owner, map);
			}
			
			//public variable access from outside class
			if(modifier == AccessModifier.PUBLIC.ordinal() && 
					!owner.equals(executableCode.getOwnerUserType())){
				publicFieldAccesses++;
			}
			//access from outside object
			if(modifier == AccessModifier.PUBLIC.ordinal()){
				if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
						&& !owner.equals(executableCode.getOwnerUserType())){
					publicFieldAccessesOtherObject++;
				}
				else {
					int accesses = getOtherObjectAccesses(expression);
					publicFieldAccessesOtherObject += accesses;
				}
			}
			
			//package variable access from outside class
			if(modifier == AccessModifier.PACKAGE.ordinal() && 
					!owner.equals(executableCode.getOwnerUserType())){
				packageFieldAccessesOther++;
			}
			//access from outside the object
			if(modifier == AccessModifier.PACKAGE.ordinal()){
				if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
						&& !owner.equals(executableCode.getOwnerUserType())){
					packageFieldAccessesOtherObject++;
				}
				else {
					int accesses = getOtherObjectAccesses(expression);
					packageFieldAccessesOtherObject += accesses;
				}
			}
		
			//protected variable access
			if(modifier == AccessModifier.PROTECTED.ordinal() &&
					!owner.equals(executableCode.getOwnerUserType())){
				//from decendant
				if(isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())){
					protectedFieldAccessesAncestor++;
				}
				//from outside class
				else if(!owner.equals(executableCode.getOwnerUserType())){
					protectedFieldAccessesOther++;
				}
			}
			
			//protected variable access from another object
			if(modifier == AccessModifier.PROTECTED.ordinal()){
				//from outside object
				if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
						&& !owner.equals(executableCode.getOwnerUserType())){
					protectedFieldAccessOtherObject++;
				}
				//from descendant
				else {
					int accesses = getOtherObjectAccesses(expression);
					protectedFieldAccessOtherObject += accesses;
				}
				
				//protected method access to ancestor from same object
				if(isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())){
					int accesses = getOtherObjectAccesses(expression);
					protectedFieldAccessesAncestorObject += (1 - accesses);
				}
			}
			
			//check for private access from within same class but different object
			if(modifier == AccessModifier.PRIVATE.ordinal() && 
					executableCode.getOwnerUserType().equals(owner)){
				int accesses = getOtherObjectAccesses(expression);
				privateFieldAccesses += accesses;
			}
		}
		
		//method invocation
		if(decl instanceof OperationDecl) {
			OperationDecl method = (OperationDecl) decl;
			
			//store information about the method access
	        if(!methodAccessMap.containsKey(method)){
	        	Set<UserType> set = new HashSet<UserType>();
	        	set.add(executableCode.getOwnerUserType());
	        	methodAccessMap.put(method, set);
	        }
	        else {
	        	Set<UserType> set = methodAccessMap.get(method);
	        	set.add(executableCode.getOwnerUserType());
	        	methodAccessMap.put(method, set);
	        }
			
			totalMethodAccesses++;
			int modifier = getAccessModifier(method.getModifier());
			methodAccesses[modifier]++;
			
			//calculate the total number of accesses from outside the object / class
			if(!owner.equals(executableCode.getOwnerUserType())){
				methodAccessesOtherClass++;
			}
			if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
					&& !owner.equals(executableCode.getOwnerUserType())){
				methodAccessesOtherObject++;
			}
			else {
				int accesses = getOtherObjectAccesses(expression);
				methodAccessesOtherObject += accesses;
			}
			
			//more in depth info about method accesses
			if(methodAccessCounts.containsKey(owner)){
				Map<OperationDecl, Integer> map = methodAccessCounts.get(owner);
				if(map.containsKey(method)){
					map.put(method, map.get(method) + 1);
				}
				else {
					map.put(method, 1);
				}
			}
			else {
				Map<OperationDecl, Integer> map = new HashMap<OperationDecl, Integer>();
				map.put(method, 1);
				methodAccessCounts.put(owner, map);
			}
			
			//public method accesses from outside class
			if(modifier == AccessModifier.PUBLIC.ordinal() && 
					!owner.equals(executableCode.getOwnerUserType())){
				publicMethodAccesses++;
			}
			//access from outside object
			if(modifier == AccessModifier.PUBLIC.ordinal()){
				if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
						&& !owner.equals(executableCode.getOwnerUserType())){
					publicMethodAccessesOtherObject++;
				}
				else {
					int accesses = getOtherObjectAccesses(expression);
					publicMethodAccessesOtherObject += accesses;
				}
			}
			
			//package method accesses from outside class
			if(modifier == AccessModifier.PACKAGE.ordinal() && 
					!owner.equals(executableCode.getOwnerUserType())){
				packageMethodAccessesOther++;
			}
			//access from outside object
			if(modifier == AccessModifier.PACKAGE.ordinal()){
				if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
						&& !owner.equals(executableCode.getOwnerUserType())){
					packageMethodAccessesOtherObject++;
				}
				else {
					int accesses = getOtherObjectAccesses(expression);
					packageMethodAccessesOtherObject += accesses;
				}
			}
		
			//protected method accesses
			if(modifier == AccessModifier.PROTECTED.ordinal() &&
					!owner.equals(executableCode.getOwnerUserType())){
				//from descendant
				if(isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())){
					protectedMethodAccessesAncestor++;
				}
				//from outside class
				else {
					protectedMethodAccessesOther++;
				}
			}
			
			//protected method accesses from another object
			if(modifier == AccessModifier.PROTECTED.ordinal()){
				//from outside object
				if(!isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())
						&& !owner.equals(executableCode.getOwnerUserType())){
					protectedFieldAccessOtherObject++;
				}
				//from descendant
				else {
					int accesses = getOtherObjectAccesses(expression);
					protectedMethodAccessOtherObject += accesses;
				}
				
				//protected method access to ancestor from same object
				if(isAncestor((ClassType)owner, (ClassType)executableCode.getOwnerUserType())){
					int accesses = getOtherObjectAccesses(expression);
					protectedMethodAccessesAncestorObject += (1 - accesses);
				}
			}
			
			//check for private access from within same class but different object
			if(modifier == AccessModifier.PRIVATE.ordinal() && 
					executableCode.getOwnerUserType().equals(owner)){
				int accesses = getOtherObjectAccesses(expression);
				privateMethodAccesses += accesses;
			}
		}
    }
    
    private jst.symtab.link.Link getNextLink(jst.symtab.link.Link link){
    	if(link instanceof IdLookup){
			return ((IdLookup)link).getQualifier();
		}
		else {
			return null;
		}
    }
    
    /*
     * Get the part of the expression that's relevant for this decl access
     * If the decl we are looking for is called a, then:
     * For "a", return a
     * For "this.a", return this.a
     * For "this.b.a", return b.a
     */
    private String getExpression(jst.symtab.link.Link link, Decl decl) {
    	SyntaxTree source = link.getSource();
    	
    	if(source == null) {
    		return null;
    	}
    	
    	while(source != null &&
    			!source.getSymbolType().getName().equals(expressionName) &&
    			!source.getSymbolType().getName().equals(lhsName) &&
    			!source.getSymbolType().getName().equals(expressionStatementName) &&
    			!source.getSymbolType().getName().equals(expressionNameName)){
    		source = source.getParent();
    	}
    	
    	if(source == null) {
    		return null;
    	}
    	
    	String wholeExpression = " " + source.getValue() + " ";
    	
    	int index = wholeExpression.indexOf(" " + decl.getSimpleName() + " ");
    	if(index == -1) {
    		return null;
    	}
    	int dotIndex = wholeExpression.lastIndexOf(".", index);
    	if(dotIndex < index && dotIndex != -1){
    		//index of the dot before the one immediately before the access we are looking for
    		int previousDot = wholeExpression.lastIndexOf(".", dotIndex - 1);
    		int endIndex = index + decl.getSimpleName().length() + 1;
    		return wholeExpression.substring(previousDot + 1, endIndex).trim();
    	}
    	else {
    		return decl.getSimpleName();
    	}
    }
    
    private int getOtherObjectAccesses(String expression) {	
    	if(expression.contains(".") && !expression.startsWith("this") && !expression.startsWith("super")){
			return 1;
		}
    	return 0;
    }
    
    private boolean isAncestor(ClassType ancestor, ClassType decendant){
    	while(decendant.getSuperclass() != null){
    		decendant = decendant.getSuperclass();
    		if(decendant.equals(ancestor)){
    			return true;
    		}
    	}
    	return false;
    }
    
    private void resolveVisibility() {
    	for(ClassType ct : fieldCanSee.keySet()){
    		Set<FieldDecl> fieldSet = fieldCanSee.get(ct);
    		for(FieldDecl field : fieldVisibleFrom.keySet()){
    			Set<ClassType> classSet = fieldVisibleFrom.get(field);
    			if(field.getModifier() == Modifier.PUBLIC){
    				fieldSet.add(field);
    				classSet.add(ct);
    			}
    			else if(field.getModifier() == Modifier.PROTECTED){
    				if(field.getOwnerPackage().equals(ct.getOwnerPackage())
    						|| isAncestor((ClassType)field.getOwnerUserType(), ct)){
    					fieldSet.add(field);
        				classSet.add(ct);
    				}
    			}
    			else if(field.getModifier() != Modifier.PRIVATE){
    				if(field.getOwnerPackage().equals(ct.getOwnerPackage())){
    					fieldSet.add(field);
        				classSet.add(ct);
    				}
    			}
    		}
    	}
    	for(ClassType ct : methodCanSee.keySet()){
    		Set<OperationDecl> methodSet = methodCanSee.get(ct);
    		for(OperationDecl method : methodVisibleFrom.keySet()){
    			Set<ClassType> classSet = methodVisibleFrom.get(method);
    			if(method.getModifier() == Modifier.PUBLIC){
    				methodSet.add(method);
    				classSet.add(ct);
    			}
    			else if(method.getModifier() == Modifier.PROTECTED){
    				if(method.getOwnerPackage().equals(ct.getOwnerPackage())
    						|| isAncestor((ClassType)method.getOwnerUserType(), ct)){
    					methodSet.add(method);
        				classSet.add(ct);
    				}
    			}
    			else if(method.getModifier() != Modifier.PRIVATE){
    				if(method.getOwnerPackage().equals(ct.getOwnerPackage())){
    					methodSet.add(method);
        				classSet.add(ct);
    				}
    			}
    		}
    	}
    }
    
    private double getAverage(Map map) {
    	int total = 0;
    	Iterator iter = map.values().iterator();
    	while(iter.hasNext()){
    		Set set = (Set)iter.next();
    		total += set.size();
    	}
    	return total / map.size();
    }
	
	public void printResults(boolean simple) {
		//resolve the visibility of fields
		resolveVisibility();
		
		System.out.println("FIELD INFORMATION\n");
		
		//print information about variables
		System.out.println("Total Field Count: " + totalFields);
		
		if(totalFields > 0) {
			System.out.println("Public: " + fieldCounts[AccessModifier.PUBLIC.ordinal()]);
			System.out.println("Protected: " + fieldCounts[AccessModifier.PROTECTED.ordinal()]);
			System.out.println("Package: " + fieldCounts[AccessModifier.PACKAGE.ordinal()]);
			System.out.println("Private: " + fieldCounts[AccessModifier.PRIVATE.ordinal()] + "\n");
			
			System.out.println("Percentage of fields visible per class: " + (getAverage(fieldCanSee) * 100 / fieldVisibleFrom.size()));
			System.out.println("Percentage of classes a field is visible to on average: " + (getAverage(fieldVisibleFrom) * 100 / fieldCanSee.size()) + "\n");
		}
		//print information about variable accesses
		System.out.println("Total Field Accesses: " + totalFieldAccesses);
		if(totalFieldAccesses > 0) {
			System.out.println("Total Accesses per Field: " + (float)totalFieldAccesses / totalFields);
			System.out.println("Total Accesses from Outside the Class: " + fieldAccessesOtherClass);
			System.out.println("Total Accesses from Outside the Object: " + fieldAccessesOtherObject);
			int minimum = getMinAccessesTotal(fieldAccessCounts);
			int maximum = getMaxAccessesTotal(fieldAccessCounts);
			System.out.println("Minimum number of accesses per field: " + minimum);
			System.out.println("Maximum number of accesses per field: " + maximum);
			System.out.println("Public: " + fieldAccesses[AccessModifier.PUBLIC.ordinal()]);
			System.out.println("Protected: " + fieldAccesses[AccessModifier.PROTECTED.ordinal()]);
			System.out.println("Package: " + fieldAccesses[AccessModifier.PACKAGE.ordinal()]);
			System.out.println("Private: " + fieldAccesses[AccessModifier.PRIVATE.ordinal()]);
			System.out.println("Accesses to public fields from outside the class: " + publicFieldAccesses);
			System.out.println("Accesses to public fields from outside the object: " + publicFieldAccessesOtherObject);
			System.out.println("Accesses to package fields from outside the class: " + packageFieldAccessesOther);
			System.out.println("Accesses to package fields from outside the object: " + packageFieldAccessesOtherObject);
			System.out.println("Accesses to protected fields from decendants: " + protectedFieldAccessesAncestor);
			System.out.println("Accesses to protected fields from another class: " + protectedFieldAccessesOther);
			System.out.println("Accesses to protected fields from outside the object: " + protectedFieldAccessOtherObject);
			System.out.println("Accesses to private fields of another object of the same class: " + privateFieldAccesses + "\n");
		}
		
		//per class info
		if(!simple) {
			for(UserType type : classFieldCounts.keySet()){
				System.out.println("Variable Counts for Class " + type.getSimpleName());
				List<Integer> list = classFieldCounts.get(type);
				int totalVars = getTotal(list);
				System.out.println("Total number of fields: " + totalVars);
				System.out.println("Public: " + list.get(AccessModifier.PUBLIC.ordinal()));
				System.out.println("Protected: " + list.get(AccessModifier.PROTECTED.ordinal()));
				System.out.println("Package: " + list.get(AccessModifier.PACKAGE.ordinal()));
				System.out.println("Private: " + list.get(AccessModifier.PRIVATE.ordinal()));

				Map<FieldDecl, Integer> accesses = fieldAccessCounts.get(type);
				int totalAccesses = getTotal(accesses.values());
				System.out.println("Total number of accesses to fields: " + totalAccesses);
				System.out.println("Average number of accesses per field: " + (double)totalAccesses / totalVars);
				int minimum = getMinAccesses(fieldAccessCounts.get(type));
				int maximum = getMaxAccesses(fieldAccessCounts.get(type));
				System.out.println("Minimum number of accesses per field: " + minimum);
				System.out.println("Maximum number of accesses per field: " + maximum + "\n");
			}
		}
		
		System.out.println("METHOD INFORMATION\n");
		
		//print information about methods
		System.out.println("Total Method Count: " + totalMethods);
		if(totalMethods > 0){
			System.out.println("Public: " + methodCounts[AccessModifier.PUBLIC.ordinal()]);
			System.out.println("Protected: " + methodCounts[AccessModifier.PROTECTED.ordinal()]);
			System.out.println("Package: " + methodCounts[AccessModifier.PACKAGE.ordinal()]);
			System.out.println("Private: " + methodCounts[AccessModifier.PRIVATE.ordinal()] + "\n");
			
			System.out.println("Percentage of methods visible per class: " + (getAverage(methodCanSee) * 100 / methodVisibleFrom.size()));
			System.out.println("Percentage of classes a methods is visible to on average: " + (getAverage(methodVisibleFrom) * 100 / methodCanSee.size()) + "\n");
		}
		//print information about method accesses
		System.out.println("Total Method Accesses: " + totalMethodAccesses);
		if(totalMethodAccesses > 0) {
			System.out.println("Total Accesses per Method: " + (float)totalMethodAccesses / totalMethods);
			System.out.println("Total Accesses from Outside the Class: " + methodAccessesOtherClass);
			System.out.println("Total Accesses from Outside the Object: " + methodAccessesOtherObject);
			int minimum = getMinAccessesTotal(methodAccessCounts);
			int maximum = getMaxAccessesTotal(methodAccessCounts);
			System.out.println("Minimum number of accesses per method: " + minimum);
			System.out.println("Maximum number of accesses per method: " + maximum);
			System.out.println("Public: " + methodAccesses[AccessModifier.PUBLIC.ordinal()]);
			System.out.println("Protected: " + methodAccesses[AccessModifier.PROTECTED.ordinal()]);
			System.out.println("Package: " + methodAccesses[AccessModifier.PACKAGE.ordinal()]);
			System.out.println("Private: " + methodAccesses[AccessModifier.PRIVATE.ordinal()]);
			System.out.println("Accesses to public methods from outside the class: " + publicMethodAccesses);
			System.out.println("Accesses to public methods from outside the object: " + publicMethodAccessesOtherObject);
			System.out.println("Accesses to package methods from outside the class: " + packageMethodAccessesOther);
			System.out.println("Accesses to package methods from outside the object: " + packageMethodAccessesOtherObject);
			System.out.println("Accesses to protected methods from decendants: " + protectedMethodAccessesAncestor);
			System.out.println("Accesses to protected methods from another class: " + protectedMethodAccessesOther);
			System.out.println("Accesses to protected methods from outside the object: " + protectedMethodAccessOtherObject);
			System.out.println("Accesses to private methods of another object of the same class: " + privateMethodAccesses + "\n");
		}
		//per class info
		if(!simple) {
			for(UserType type : methodClassCounts.keySet()){
				System.out.println("Method Counts for Class " + type.getSimpleName());
				List<Integer> list = methodClassCounts.get(type);
				int numMethods = getTotal(list);
				System.out.println("Total number of methods: " + numMethods);
				System.out.println("Public: " + list.get(AccessModifier.PUBLIC.ordinal()));
				System.out.println("Protected: " + list.get(AccessModifier.PROTECTED.ordinal()));
				System.out.println("Package: " + list.get(AccessModifier.PACKAGE.ordinal()));
				System.out.println("Private: " + list.get(AccessModifier.PRIVATE.ordinal()));

				Map<OperationDecl, Integer> accesses = methodAccessCounts.get(type);
				for(OperationDecl dec : accesses.keySet()){
					System.out.println(dec.getSimpleName() + " " + accesses.get(dec));
				}
				int totalAccesses = getTotal(accesses.values());
				if(totalAccesses > 0) {
					System.out.println("Total number of accesses to methods: " + totalAccesses);
					System.out.println("Average number of accesses per method: " + (double)totalAccesses / numMethods);
					int minimum = getMinAccesses(methodAccessCounts.get(type));
					int maximum = getMaxAccesses(methodAccessCounts.get(type));
					System.out.println("Minimum number of accesses per method: " + minimum);
					System.out.println("Maximum number of accesses per method: " + maximum + "\n");
				}
			}
		}
		double classEncapsulationAccesses = privateMethodAccesses + privateFieldAccesses;
		double objectEncapsulationAccesses = protectedFieldAccessesAncestorObject + 
				protectedMethodAccessesAncestorObject;
		double totalAccesses = totalMethodAccesses + totalFieldAccesses;
		
		if(totalAccesses > 0) {
		
			//less than 10 percent different between the two
			if(Math.abs(classEncapsulationAccesses - objectEncapsulationAccesses) / (classEncapsulationAccesses +  objectEncapsulationAccesses ) < 0.25){
				System.out.println("You practice a mixture of class and object encapsulation with "
						+ (classEncapsulationAccesses * 100 / totalAccesses) + "% of member accesses indicating class encapsulation and "
						+ (objectEncapsulationAccesses * 100 / totalAccesses) + "% indicating object encapsulation.");
			}
			//more than 90 percent difference
			else if(Math.abs(classEncapsulationAccesses - objectEncapsulationAccesses) / (classEncapsulationAccesses +  objectEncapsulationAccesses ) > 0.5){
				if(objectEncapsulationAccesses > classEncapsulationAccesses){
					System.out.println("You practice mostly object encapsulation with "
						+ (classEncapsulationAccesses * 100 / totalAccesses) + "% of member accesses indicating class encapsulation and "
						+ (objectEncapsulationAccesses * 100 / totalAccesses) + "% indicating object encapsulation.");
				}
				else {
					System.out.println("You practice mostly class encapsulation with "
							+ (classEncapsulationAccesses * 100 / totalAccesses) + "% of member accesses indicating class encapsulation and "
							+ (objectEncapsulationAccesses * 100 / totalAccesses) + "% indicating object encapsulation.");
				}
			}
			else if(objectEncapsulationAccesses > classEncapsulationAccesses){
				System.out.println("You practice more object than class encapsulation with "
						+ (classEncapsulationAccesses * 100 / totalAccesses) + "% of member accesses indicating class encapsulation and "
						+ (objectEncapsulationAccesses * 100 / totalAccesses) + "% indicating object encapsulation.");
			}
			else {
				System.out.println("You practice more class than object encapsulation with "
						+ (classEncapsulationAccesses * 100 / totalAccesses) + "% of member accesses indicating class encapsulation and "
						+ (objectEncapsulationAccesses * 100 / totalAccesses) + "% indicating object encapsulation.");
			}
		}
	}
	
	private int getTotal(Collection<Integer> collection){
		int sum = 0;
		for(Integer i : collection) {
			sum += i;
		}
		return sum;
	}
	
	private int getMinAccessesTotal(Map map){
		int minimum = -1;
		Iterator iter = map.keySet().iterator();
		while(iter.hasNext()){
			UserType ut = (UserType) iter.next();
			Map accessMap = (Map)map.get(ut);
			Iterator iter2 = accessMap.values().iterator();
			while(iter2.hasNext()){
				Integer i = (Integer) iter2.next();
				if(i < minimum || minimum == -1){
					minimum = i;
				}
			}
		}
		return minimum;
	}
	
	private int getMaxAccessesTotal(Map map){
		int maximum = -1;
		Iterator iter = map.keySet().iterator();
		while(iter.hasNext()){
			UserType ut = (UserType) iter.next();
			Map accessMap = (Map)map.get(ut);
			Iterator iter2 = accessMap.values().iterator();
			while(iter2.hasNext()){
				Integer i = (Integer) iter2.next();
				if(i > maximum){
					maximum = i;
				}
			}
		}
		return maximum;
	}
	
	private int getMinAccesses(Map map){
		int minimum = -1;
		Iterator iter = map.values().iterator();
		while(iter.hasNext()){
			Integer i = (Integer)iter.next();
			if(i < minimum || minimum == -1){
				minimum = i;
			}
		}
		return minimum;
	}
	
	private int getMaxAccesses(Map map){
		int maximum = -1;
		Iterator iter = map.values().iterator();
		while(iter.hasNext()){
			Integer i = (Integer)iter.next();
			if(i > maximum){
				maximum = i;
			}
		}
		return maximum;
	}
	
}
