package encapsulation;

import jst.symtab.*;
import jst.symtab.visitor.ImportResolver;
import jst.symtab.visitor.MemberTypeResolver;
import jst.symtab.visitor.ReferenceResolver;
import jst.symtab.visitor.SuperInterfaceResolver;
import jst.symtab.visitor.SuperclassResolver;
import jst.symtab.visitor.XMLWriter;
import jst.parsetree.*;

import java.io.*;
import java.util.SortedSet;
import java.util.TreeSet;


/** The main entrypoint for the application.
 *  It loads marked-up parse trees that were produced by a Java parser, and builds a symbol table.
 *  Symbol references are looked up in the symbol table so that all identifiers are fully cross-
 *  referenced.  Classes not found in source files (parse trees) are expected to be available in
 *  ".class" files examined introspectively.  The cross-referenced symbol table and parse trees
 *  are saved as an XML file.
 */
public class Main {

    public static final String USAGE =	"Usage: java encapsulation.Main [-c | -s] <source dir> [<source dir> ...] <new source dir>";

	/** Run the application.
     *  @param args An array of file or directory names containing parse trees.
     *         Any file ending in ".xml" is assumed to be a parse tree.
     */
    public static void main(String[] args) {
    	//create the model from the file
		Main app = new Main();
		
		//check if the first argument indicates simple or complex output requirements
		//simple is the default if the argument is not specified
		int startIndex = 0;
		boolean simple = true;
		//simple output
		if(args[0].equals("-s")){
			startIndex = 1;
		}
		//complex output
		else if(args[0].equals("-c")){
			simple = false;
			startIndex = 1;
		}

		// read source files and cross reference
		app.run(args, startIndex);
		
		//collect encapsulation data
		EncapsulationAnalysisVisitor visitor = new EncapsulationAnalysisVisitor();
		PackageDecl pack = PackageDecl.getDefaultPackage();
		pack.accept(visitor);
		//print results
		visitor.printResults(simple);
    }

	public static void error(String message) {
    	System.err.println("Error--> " + message);
    }

	/** Run the application.
     *  Load each parse tree, populate the symbol table, lookup references, then emit xml.
     */
    protected void run(String[] args, int startIndex) {
        if (args.length <= startIndex) {
            System.err.println(USAGE);
            return;
        }

		readSources(args, startIndex);
        resolve();
        //save();
    }

	protected void readSources(String[] args, int startIndex) {
		System.err.println("Reading parse trees...");
		for (int i = startIndex; i < args.length; i++) // for each dir/file on the command line
			readSource(new File(args[i]));
	}

    /** Walk directory recursively, looking for xml files to read.
     *  Anything ending in ".xml" is assumed to contain a parse tree.
     *  Trees are loaded into the parseTrees array.
     */
    public void readSource(File file) {
        if (!file.exists()) {
            System.err.println("File not found: " + file);
            return;
        }

		// If this is a directory, walk each file/dir in that directory
		if (file.isDirectory()) {
			String files[] = file.list();
			for (int i = 0; i < files.length; i++) {
				File next = new File(file, files[i]);
				readSource(next);
			}
            return;
		}

		// Is this an xml file?
		if ((file.getName().length() <= 4) ||
			!file.getName().substring(file.getName().length() - 4).equals(".xml")) {
            return; // Not a ".xml" file.  Ignore it.
        }

		// Read the parse tree into memory.
        System.err.println("Reading " + file.getAbsolutePath());
		ParseTree tree = new ParseTree(file);
        addParseTree(tree);
    }

    public void addParseTree(ParseTree tree) {
        SourceFile sourceFile = new SourceFile(tree);
        // sourceFiles.add(sourceFile);

		// Find declarations in the parse tree.
        DeclarationVisitor visitor = new DeclarationVisitor(sourceFile);
        tree.getRoot().accept(visitor);
    }

    public void resolve() {
		System.err.println("Resolving...");
    	SortedSet<String> messages = new TreeSet<String>();

		ImportResolver importResolver = new ImportResolver();
		PackageDecl.getDefaultPackage().accept(importResolver);

		// for (JSTMessage mess: importLinker.getMessages())
		for (Object mess: importResolver.getMessages())
			messages.add(mess.toString());

		SuperclassResolver superclassResolver = new SuperclassResolver();
		PackageDecl.getDefaultPackage().accept(superclassResolver);

		for (Object mess: superclassResolver.getMessages())
			messages.add(mess.toString());
		
		SuperInterfaceResolver superInterfaceResolver = new SuperInterfaceResolver();
		PackageDecl.getDefaultPackage().accept(superInterfaceResolver);

		for (Object mess: superInterfaceResolver.getMessages())
			messages.add(mess.toString());

		MemberTypeResolver typeResolver = new MemberTypeResolver();
		PackageDecl.getDefaultPackage().accept(typeResolver);

		for (Object mess: typeResolver.getMessages())
			messages.add(mess.toString());

		System.err.println("Resolving refs...");
		ReferenceResolver refResolver = new ReferenceResolver();
		PackageDecl.getDefaultPackage().accept(refResolver);

		for (Object mess: refResolver.getMessages())
			messages.add(mess.toString());

		if (messages.isEmpty())
			System.err.println("OK");
		else {
			System.err.println(messages.size() + " errors:");

			for (String message: messages)
				System.err.println("  " + message);
		}
    }

	protected void save() {
		System.err.println("Writing XML file...");
		PrintWriter out = new PrintWriter(new BufferedWriter(new OutputStreamWriter(System.out)));

		out.println("<?xml version='1.0' encoding='iso-8859-1'?>");
		out.println("<!DOCTYPE program SYSTEM 'jst.dtd'>");
		out.println("<program>");
        out.println("<symbols>");
        XMLWriter xml = new XMLWriter(out);
        PackageDecl.getDefaultPackage().accept(xml);
        out.println("</symbols>");
    	out.println("</program>");

		out.close();
    }
}


