
package jst.symtab;

import java.util.*;
import java.io.*;
import java.lang.reflect.Modifier;

import jst.parsetree.*;
import jst.symtab.visitor.*;
import jst.symtab.link.*;

/**
 * A block of statements delimited by curly braces.
 */
public class Block extends ExecutableCode implements Scope {

    protected int staticModifier;

    /**
     *@link aggregation
     *      @associates <{Block}>
     * @supplierCardinality 0..*
     * @supplierRole blocks
     */
    protected List<Block> blocks;

    /**
     *@link aggregation
     *      @associates <{UserType}>
     * @supplierCardinality 0..*
     * @supplierRole types*/
    protected Map<String, UserType> types; // Simple name ==> UserType

    /**
     *@link aggregation
     *      @associates <{LocalVariableDecl}>
     * @supplierCardinality 0..*
     * @supplierRole variables*/
    protected Map<String, LocalVariableDecl> variables;

	/**
     * Internal constructor.
     */
	private Block(Decl theOwner, String theSimpleName, JstParseTreeBranch theSource) {
    	super(theOwner, theSimpleName, theSource);
    }

	/**
     * Construct an outermost block.
     */
	public Block(OperationDecl ownerOperation, JstParseTreeBranch theSource) {
    	this(ownerOperation, "@BODY", theSource);

    	ownerOperation.setBody(this);
    }

	/**
     * Construct a block nested inside another.
     */
	public Block(Block ownerBlock, JstParseTreeBranch theSource) {
    	this(ownerBlock, ownerBlock.generateName(), theSource);

    	ownerBlock.addBlock(this);
    }

    /**
     * Construct an initializer block in a class.
     */
    public Block(ClassType ownerClass, JstParseTreeBranch theSource) {
        this(ownerClass, ownerClass.generateName(), theSource);

        ownerClass.addInitializerBlock(this);
    }
    
    protected Block(Decl newOwner, Block that, Map<TypeParameter, ActualType> swaps) {
		super(newOwner, that, swaps);

		for (Link thatLink : references)
			references.add(thatLink.copy(this));

		for (Block thatBlock : that.blocks)
			thatBlock.copy(this, swaps);

		for (UserType thatUserType : that.types.values())
			thatUserType.copy(this, swaps);

		for (LocalVariableDecl thatVariable : that.variables.values())
			thatVariable.copy(this, swaps);
	}

    public Block(OperationDecl newOwnerOperation, Block that, Map<TypeParameter, ActualType> swaps) {
		this((Decl) newOwnerOperation, that, swaps);

    	newOwnerOperation.setBody(this);
	}

    public Block(Block newOwnerBlock, Block that, Map<TypeParameter, ActualType> swaps) {
		this((Decl) newOwnerBlock, that, swaps);

    	newOwnerBlock.addBlock(this);
	}

    public Block(ClassType newOwnerClass, Block that, Map<TypeParameter, ActualType> swaps) {
		this((Decl) newOwnerClass, that, swaps);

        newOwnerClass.addInitializerBlock(this);
	}

	protected void init() {
		super.init();

        blocks = new ArrayList<Block>();
        types = new HashMap<String, UserType>();
        variables = new HashMap<String, LocalVariableDecl>();
	}

	public void copy(OperationDecl newOwnerOperation, Map<TypeParameter, ActualType> swaps) {
		new Block(newOwnerOperation, this, swaps);
	}
	
	public void copy(Block newOwnerBlock, Map<TypeParameter, ActualType> swaps) {
		new Block(newOwnerBlock, this, swaps);
	}

	public void copy(ClassType newOwnerClass, Map<TypeParameter, ActualType> swaps) {
		new Block(newOwnerClass, this, swaps);
	}

 	public String getCanonicalName() {
       return "BLK_" + getOwnerScope().getCanonicalName().substring(4) + "." + simpleName;
    }

	public Scope asScope() {
		return this;
	}

    // Used for telling if the block is static.
    public int getModifier() {
        return staticModifier;
    }

    public void setModifier(int modifier) {
        staticModifier = modifier;
    }

	public void addBlock(Block subBlock) {
    	blocks.add(subBlock);
    }

	public List getBlocks() {
    	return blocks;
    }

    public boolean addType(UserType type) {
        types.put(type.getSimpleName(), type);
        return true;
    }

    public Set getTypes() {
        return new HashSet(types.values());
    }

    public TypeDecl lookupType(boolean widen, String typeName, ReferenceType looker) {
		TypeDecl result = (UserType) types.get(typeName);

		if (result == null && widen)
			result = getOwnerScope().lookupType(true, typeName, looker);

		return result;
	}

	public void addVariable(LocalVariableDecl var) {
    	variables.put(var.getSimpleName(), var);
    }

    public Set getVariables() {
        return new HashSet(variables.values());
    }

	public VariableDecl lookupVariable(boolean widen, String name, ReferenceType looker) {
    	VariableDecl result = (VariableDecl) variables.get(name);

		if (result == null && widen)
			result = getOwnerScope().lookupVariable(true, name, looker);

		return result;
    }

	public MethodDecl lookupMethod(boolean widen, String methodName, List params, ReferenceType looker) {
        return widen? getOwnerScope().lookupMethod(true, methodName, params, looker) : null;
    }
	
	protected Scope getScope() {
		return this;
	}

    public void accept(ModelVisitor visitor) {
        visitor.visitBlock(this);
    }
    
}
