package jst.symtab;

import java.util.*;

import parser.parsetree.*;
import jst.parsetree.*;
import jst.symtab.link.*;


public class Builder {
	
	protected Decl currentDecl;
	
	protected Link currentLink;

	protected Stack<Link> nestedLinks;
	
	protected PackageDecl pkg;

	
	public Builder(SourceFile src) {
		nestedLinks = new Stack<Link>();
		pkg = PackageDecl.getDefaultPackage();
		pushDecl(src);

		// Every file effectively imports java.lang.
		currentLink = new PackageOrTypeLookup(getScopeLink(), id("java"));
		currentLink = new PackageOrTypeLookup(getScopeLink(), id("lang"));
		importOnDemand(false);
	}

	// Yuk.  We need Jst parse trees, but visitors that call this builder
	// are aware only of generic syntax trees.
	// For now, encapsulate the cast here.
	protected JstParseTreeBranch jstTree(SyntaxTreeBranch tree) {
		return (JstParseTreeBranch) tree;
	}
	
	protected void pushDecl(Decl decl) {
		assert decl != null;
		assert currentLink == null : currentLink;
		
		currentDecl = decl;
	}
	
	protected void popDecl() {
		currentDecl = getDecl().getOwnerDecl();
		assert currentDecl != null;
	}
	
	protected void pushLink() {
		nestedLinks.push(getLink());
		currentLink = null;
	}
	
	protected void popLink() {
		currentLink = nestedLinks.pop();
	}

	protected Decl getDecl() {
		assert currentDecl != null;

		return currentDecl;
	}

	public String getDeclName() {
		return getDecl().getSimpleName();
	}

	protected SourceFile getSourceFile() {
		return (SourceFile) getDecl();
	}
	
	protected UserType getUserType() {
		return (UserType) getDecl();
	}

	protected GenericClassType getClassType() {
		return (GenericClassType) getDecl();
	}

	protected GenericInterfaceType getInterfaceType() {
		return (GenericInterfaceType) getDecl();
	}

	protected TypeParameter getTypeParameter() {
		return (TypeParameter) getDecl();
	}

	protected Block getBlock() {
		return (Block) getDecl();
	}

	protected CatchBlock getCatchBlock() {
		return (CatchBlock) getDecl();
	}
	protected ExecutableCode getExecutableCode() {
		if(getDecl() instanceof ExecutableCode){
			return (ExecutableCode) getDecl();
		}
		else {
			Decl decl = getDecl();
			while(decl != null && !(decl instanceof ExecutableCode)){
				decl = decl.getOwnerDecl();
			}
			return (ExecutableCode)decl;
		}
	}

	protected OperationDecl getOperationDecl() {
		return (OperationDecl) getDecl();
	}

	protected TypedDecl getTypedDecl() {
		return (TypedDecl) getDecl();
	}

	protected FieldDecl getFieldDecl() {
		return (FieldDecl) getDecl();
	}

	protected Link getLink() {
		assert currentLink != null;

		return currentLink;
	}
	
	protected ScopeLink getScopeLink() {
		if (currentLink != null)
			return (ScopeLink) getLink();

		return new KnownDecl(getDecl());
	}
	
	protected TypeLink getTypeLink() {
		return (TypeLink) getLink();
	}
	
	protected ActualTypeLink getActualTypeLink() {
		return (ActualTypeLink) getLink();
	}

	protected BuiltInTypeLink getBuiltInTypeLink() {
		return (BuiltInTypeLink) getLink();
	}

	protected TypeLookup getTypeLookup() {
		return (TypeLookup) getLink();
	}

	protected UserTypeLookup getUserTypeLookup() {
		return (UserTypeLookup) getLink();
	}

	protected ClassTypeLookup getClassTypeLookup() {
		return (ClassTypeLookup) getLink();
	}

	protected InterfaceTypeLookup getInterfaceTypeLookup() {
		return (InterfaceTypeLookup) getLink();
	}

	protected ReferenceTypeLink getReferenceTypeLink() {
		return (ReferenceTypeLink) getLink();
	}

	protected PackageOrTypeLookup getPackageOrTypeLookup() {
		return (PackageOrTypeLookup) getLink();
	}

	protected WildcardLink getWildcardLink() {
		return (WildcardLink) getLink();
	}

	protected OperationLookup getOperationLookup() {
		return (OperationLookup) getLink();
	}
	
	protected Identifier id(SyntaxTreeLeaf idToken) {
		return new SourceIdentifier(idToken);
	}

	protected Identifier id(String idString) {
		return new StringIdentifier(idString, 0, 0);
	}	

	public void lookupPackageOrType(SyntaxTreeLeaf idToken) {
		currentLink = new PackageOrTypeLookup(getScopeLink(), id(idToken));
	}
	
	public void lookupAmbiguous(SyntaxTreeLeaf idToken) {
		currentLink = new AmbiguousLookup(getScopeLink(), id(idToken));
	}

	public void lookupType(SyntaxTreeLeaf idToken) {
		currentLink = new TypeLookup(getScopeLink(), id(idToken));
	}

	public void lookupUserType(SyntaxTreeLeaf idToken) {
		currentLink = new UserTypeLookup(getScopeLink(), id(idToken));
	}

	public void lookupClassType(SyntaxTreeLeaf idToken) {
		currentLink = new ClassTypeLookup(getScopeLink(), id(idToken));
	}

	public void lookupClassType(String idString) {
		currentLink = new ClassTypeLookup(getScopeLink(), id(idString));
	}

	public void lookupInterfaceType(SyntaxTreeLeaf idToken) {
		currentLink = new InterfaceTypeLookup(getScopeLink(), id(idToken));
	}

	public void lookupVariable(SyntaxTreeLeaf idToken) {
		currentLink = new VariableLookup(getScopeLink(), id(idToken));
	}

	/*
	public void lookupConstructor(SyntaxTreeLeaf idToken) {
		currentLink = new ConstructorLookup(getScopeLink(), id(idToken));
	}
	*/
	public void lookupConstructor() {
		currentLink = new ConstructorLookup(getUserTypeLookup());
	}

	public void lookupMethod(SyntaxTreeLeaf idToken) {
		currentLink = new MethodLookup(getScopeLink(), id(idToken));
	}

	public void lookupArray() {
		currentLink = new ArrayLink(getTypeLink());
	}

	public void lookupWildcard(SyntaxTreeLeaf hookToken) {
		assert currentLink == null;

		currentLink = new WildcardLink(id(hookToken));
	}
	
	public void startTypeArgument() {
		assert getTypeLookup() != null;

		pushLink();
	}

	public void endTypeArgument() {

		ActualTypeLink arg = getActualTypeLink();

		popLink();

		TypeLookup type = getTypeLookup();
		type.addTypeArg(arg);
	}
	
	public void startOpTypeArgument() {
		assert getOperationLookup() != null;

		pushLink();
	}

	public void endOpTypeArgument() {

		ReferenceTypeLink arg = getReferenceTypeLink();

		popLink();

		OperationLookup op = getOperationLookup();
		op.addTypeArgument(arg);
	}
	
	public void startBound() {
		assert getWildcardLink() != null;

		pushLink();
	}

	public void endBound(boolean upper) {

		ReferenceTypeLink bound = getReferenceTypeLink();

		popLink();

		WildcardLink wild = getWildcardLink();
		wild.setBound(upper, bound);
	}
	
	public void startArg() {
		assert getOperationLookup() != null;

		pushLink();
	}

	public void endArg() {

		TypeLink type = getTypeLink();

		popLink();

		OperationLookup op = getOperationLookup();
		op.addArgument(type);
	}

	public void voidType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getVoid(id(idToken));
	}

	public void booleanType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getBoolean(id(idToken));
	}

	public void booleanType() {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getBoolean(id("boolean"));
	}

	public void byteType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getByte(id(idToken));
	}

	public void shortType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getShort(id(idToken));
	}

	public void intType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getInt(id(idToken));
	}

	public void longType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getLong(id(idToken));
	}

	public void charType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getChar(id(idToken));
	}

	public void floatType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getFloat(id(idToken));
	}

	public void doubleType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getDouble(id(idToken));
	}

	/*
	public void nullType(SyntaxTreeLeaf idToken) {
		assert currentLink == null;

		currentLink = BuiltInTypeLink.getNull(id(idToken));
	}
	*/

	public void literal(SyntaxTreeLeaf literalToken) {
		assert currentLink == null;

		currentLink = new BuiltInTypeLink(literalToken);
	}

	public void unaryPromotion() {
		TypeLink type = getTypeLink();

		if (type instanceof BuiltInTypeLink) {
			BuiltInTypeLink builtIn = (BuiltInTypeLink) type;
			builtIn.unaryNumericPromotion();
		}
	}
	
	public void start2ndExpression() {
		assert getTypeLink() != null;

		pushLink();
	}
	
	public void end2ndExpression() {
		assert getTypeLink() != null;

		addReference();
		popLink();
	}

	public void binaryPromotion() {

		TypeLink secondType = getTypeLink();
		end2ndExpression();


		TypeLink firstType = getTypeLink();

		if (firstType instanceof BuiltInTypeLink && secondType instanceof BuiltInTypeLink) {
			BuiltInTypeLink first = (BuiltInTypeLink) firstType;
			BuiltInTypeLink second = (BuiltInTypeLink) secondType;
			first.binaryNumericPromotion(second);
		}
	}

	public void binaryPromotionOrStringConcatenation() {

		// If either type is not a primitive, it must be a String,
		// & we're really doing concatenation.

		Link secondLink = getLink();
		end2ndExpression();

		Link firstLink = getLink();

		if (firstLink instanceof BuiltInTypeLink && secondLink instanceof BuiltInTypeLink) {
			BuiltInTypeLink second = (BuiltInTypeLink) secondLink;
			BuiltInTypeLink first = (BuiltInTypeLink) firstLink;

			first.binaryNumericPromotion(second);
			return;
		}

		// FIX THIS.  Need to defer promotion until runtime.
		// We end up here for any variable, even if it is an int, because the lookup has yet to happen.
	
		addReference();

		currentLink = new PackageLookup(getScopeLink(), id("java"));
		currentLink = new PackageLookup(getScopeLink(), id("lang"));
		currentLink = new ClassTypeLookup(getScopeLink(), id("String"));
	}

	public void binaryPromotionOrBoolean() {

		// If both types are bools, result is a bool.

		Link secondLink = getLink();
		end2ndExpression();

		Link firstLink = getLink();

		if (firstLink instanceof BuiltInTypeLink && secondLink instanceof BuiltInTypeLink) {
			BuiltInTypeLink second = (BuiltInTypeLink) secondLink;
			BuiltInTypeLink first = (BuiltInTypeLink) firstLink;

			if (first.isBoolean() && second.isBoolean())
				; // leave firstType as is.
			else
				first.binaryNumericPromotion(second);
		}
	}

	public void conditionalExpression() {

		TypeLink secondType = getTypeLink();
		end2ndExpression();

		TypeLink firstType = getTypeLink();
		addReference();

		currentLink = new CombinedTypeLink(firstType, secondType);
	}

	public void dotClass(SyntaxTreeLeaf classToken) {
		assert currentLink == null;

		currentLink = new PackageOrTypeLookup(getScopeLink(), id("java"));
		currentLink = new PackageOrTypeLookup(getScopeLink(), id("lang"));
		currentLink = new ClassTypeLookup(getScopeLink(), id("Class"));
	}

	public void lookupThis(SyntaxTreeLeaf thisToken) {
		assert currentLink == null;

		// currentLink = new KnownScope(getDecl().getOwnerUserType());
		currentLink = new ClassTypeLookup(getScopeLink(), id(thisToken));
	}

	public void lookupOuterThis(SyntaxTreeLeaf thisToken) {
		assert currentLink != null;

		getClassTypeLookup().setIsInstance();
	}

	public void lookupSuper(SyntaxTreeLeaf superToken) {
		assert currentLink == null;

		currentLink = new ClassTypeLookup(getScopeLink(), id(superToken));
	}

	public void lookupOuterSuper(SyntaxTreeLeaf superToken) {
		assert currentLink != null;

		// currentLink = getClassTypeLookup().getSuperclassRef();
		// FIX ME!!!
		// Need to find outer class statically, then its superclass link.
		currentLink = null;
	}

	public void addArrayAccess() {
		assert currentLink != null;

		currentLink = new ArrayAccessLink(getTypeLink());
	}

	public void addPackageName(String name, SyntaxTreeBranch source) {

		pkg = pkg.addPackage(name, jstTree(source));

		assert pkg != null;
	}

	public void setPackage() {
		assert pkg != null;

		getSourceFile().setPackage(pkg);
		pkg = null;
	}

	/* Not needed - Default pkgs is assumed in constructor.
	public void noPackage() {
		assert pkg == null;

		getSourceFile().setPackage(PackageDecl.getDefaultPackage());
	}
	*/

	public void importType(boolean isStatic) {
		getSourceFile().addImportTypeRef(isStatic, getUserTypeLookup());
		currentLink = null;
	}

	public void importOnDemand(boolean isStatic) {
		assert getPackageOrTypeLookup() != null;

		getSourceFile().addImportOnDemandRef(isStatic, getPackageOrTypeLookup());
		currentLink = null;
	}

	public void startClass(String className, SyntaxTreeBranch normalClassDeclaration) {
		if (getDecl() instanceof SourceFile)
			pushDecl(new GenericClassType(getSourceFile(), className, jstTree(normalClassDeclaration)));
		
		else if (getDecl() instanceof UserType)
			pushDecl(new GenericClassType(getUserType(), className, jstTree(normalClassDeclaration)));
		
		else if (getDecl() instanceof Block)
			pushDecl(new GenericClassType(getBlock(), className, jstTree(normalClassDeclaration)));
		
		else if (getDecl() instanceof FieldInitializer) {
			FieldInitializer initializer = (FieldInitializer) getDecl();
			pushDecl(new GenericClassType(initializer, className, jstTree(normalClassDeclaration)));
		}
		else
			assert false : "Invalid scope for declaring class: " + getDecl();

		// What about abominable inner class in a field initializer???
	}

	public void endClass() {
		assert getClassType() != null;
		popDecl();
	}

	/*
	// Ugly Java anonymous inner class syntax requires us to find constructor in class we can never lookup.
	public void endClassAndLookupConstructor() {
	}
	*/

	public void startTypeParameter(String paramName, SyntaxTreeBranch typeParameter) {
		// Ugly casting here.

		if (getDecl() instanceof GenericClassType)
			pushDecl(new TypeParameter((GenericClassType) getDecl(), paramName, jstTree(typeParameter)));
		else if (getDecl() instanceof GenericInterfaceType)
			pushDecl(new TypeParameter((GenericInterfaceType) getDecl(), paramName, jstTree(typeParameter)));
		//else
		//	pushDecl(new TypeParameter(getOperationDecl(), paramName, jstTree(typeParameter)));
	}

	public void endTypeParameter() {
		popDecl();
	}

	public void setTypeBound() {
		getTypeParameter().addBound(getTypeLookup());
		currentLink = null;
	}

	public void startEnum(String enumName, SyntaxTreeBranch enumDeclaration) {
		startClass(enumName, enumDeclaration); // Fix this.

		assert currentLink == null;

		currentLink = new PackageOrTypeLookup(getScopeLink(), id("java"));
		currentLink = new PackageOrTypeLookup(getScopeLink(), id("lang"));
		currentLink = new ClassTypeLookup(getScopeLink(), id("Enum"));

		getClassType().setSuperclassRef(getClassTypeLookup());
		currentLink = null;
	}

	public void endEnum() {
		endClass(); // Fix this.
	}

	public void startInterface(String interfaceName, SyntaxTreeBranch normalInterfaceDeclaration) {
		if (getDecl() instanceof SourceFile)
			pushDecl(new GenericInterfaceType(getSourceFile(), interfaceName, jstTree(normalInterfaceDeclaration)));
		
		else if (getDecl() instanceof UserType)
			pushDecl(new GenericInterfaceType(getUserType(), interfaceName, jstTree(normalInterfaceDeclaration)));
		
		else if (getDecl() instanceof Block)
			pushDecl(new GenericInterfaceType(getBlock(), interfaceName, jstTree(normalInterfaceDeclaration)));
		
		else
			assert false : "invalid scope for declaring class";
	}

	public void endInterface() {
		assert getInterfaceType() != null;
		popDecl();
	}

	public void addModifier(int modifier) {
		getDecl().setModifier(getDecl().getModifier() | modifier);
	}

	public void zuper() {
		getClassType().setSuperclassRef(getClassTypeLookup());
		currentLink = null;
	}

	// When we have an anonymous superclass, we know it inherits from either a class or interface,
	// but can't tell which until we resolve the reference.
	public void maybeSuper() {
		getClassType().setSuperclassRef(getUserTypeLookup());
		currentLink = null;
	}

	public void noSuper() {
		assert currentLink == null;

		currentLink = new PackageOrTypeLookup(getScopeLink(), id("java"));
		currentLink = new PackageOrTypeLookup(getScopeLink(), id("lang"));
		currentLink = new ClassTypeLookup(getScopeLink(), id("Object"));

		getClassType().setSuperclassRef(getClassTypeLookup());
		currentLink = null;
	}

	public void superInterface() {
		getUserType().addSuperInterfaceRef(getInterfaceTypeLookup());
		currentLink = null;
	}

	public void startField(String name, SyntaxTreeBranch variableDeclaratorId) {
		pushDecl(new FieldDecl(getUserType(), name, jstTree(variableDeclaratorId)));
	}
	
	public void endField() {
		popDecl();
	}

	public void startLocalVariable(String name, SyntaxTreeBranch variableDeclaratorId) {
		pushDecl(new LocalVariableDecl(getBlock(), name, jstTree(variableDeclaratorId)));
	}
	
	public void endLocalVariable() {
		popDecl();
	}

	public void startParameter(String name, SyntaxTreeBranch variableDeclaratorId) {
		if (getDecl() instanceof OperationDecl)
			pushDecl(new ParameterDecl(getOperationDecl(), name, jstTree(variableDeclaratorId)));
		else
			pushDecl(new ParameterDecl(getCatchBlock(), name, jstTree(variableDeclaratorId)));
	}
	
	public void endParameter() {
		popDecl();
	}

	public void setType() {
		getTypedDecl().setTypeRef(getTypeLink());
		currentLink = null;
	}

	public void startMethod(String methodName, SyntaxTreeBranch methodDeclaration) {
		pushDecl(new MethodDecl(getUserType(), methodName, jstTree(methodDeclaration)));
	}

	public void endMethod() {
		popDecl();
	}
	
	public void addThrowRef() {
		getOperationDecl().addThrowRef(getTypeLookup());
	}
	
	public void startConstructor(SyntaxTreeBranch constructorDeclaration) {
		pushDecl(new ConstructorDecl(getClassType(), jstTree(constructorDeclaration)));
	}	
	
	public void endConstructor() {
		popDecl();
	}
	
	public void startBlock(SyntaxTreeBranch blockTree) {
		if (getDecl() instanceof ClassType)
			pushDecl(new Block(getClassType(), jstTree(blockTree)));
		else if (getDecl() instanceof OperationDecl)
			pushDecl(new Block(getOperationDecl(), jstTree(blockTree)));
		else if (getDecl() instanceof Block)
			pushDecl(new Block(getBlock(), jstTree(blockTree)));
		else
			assert false : "Invalid parent for a block: " + getDecl();
	}
	
	public void endBlock() {
		assert getBlock() != null;

		// addReference();

		popDecl();
	}
	
	public void startCatchBlock(SyntaxTreeBranch catchBlock) {
		pushDecl(new CatchBlock(getBlock(), jstTree(catchBlock)));
	}

	public void endCatchBlock() {
		assert getCatchBlock() != null;

		popDecl();
	}

	public void addReference() {
		if (currentLink == null)
			return;

		getExecutableCode().addReference(getLink());
		
		/*
		if (nestedLinks.isEmpty())
			currentLink = null;
		else
			popLink();
		*/
		currentLink = null;
	}
	
	public void startFieldInitializer(SyntaxTreeBranch initializer) {
		pushDecl(new FieldInitializer(getFieldDecl(), jstTree(initializer)));
	}
	
	public void endFieldInitializer() {
		assert getDecl() instanceof FieldInitializer;

		popDecl();
	}
}
