
package jst.symtab;

import java.util.*;
import java.io.*;
import jst.parsetree.*;
import java.lang.reflect.Modifier;
import jst.symtab.visitor.ModelVisitor;

// Packages form a true hierarchy, with the default package at the root.
// The name of the default package is artificial.
// Name lookups do not search up through packages, so the ownerScope of a
// package is always null.
public class PackageDecl extends Decl implements ImportableScope {

	static private PackageDecl defaultPackage = new PackageDecl(null, "@DEFAULT", null);

	static public PackageDecl getDefaultPackage() {
        return defaultPackage;
    }

	protected Set allSources; // Nonterminals.  Packages are repeatedly declared.

    /**
     *@link aggregation
     *      @associates <{PackageDecl}>
     * @supplierCardinality 0..*
     * @supplierRole subPackages*/
    protected Map subPackages; // Simple name ==> PackageDecl

    /**
     * @link association
     *      @associates <{SourceFile}>
     * @supplierCardinality 0..*
     * @supplierRole sourceFiles*/
    protected Set sourceFiles; // SourceFiles

    /**
     *@link aggregation
     *      @associates <{UserType}>
     * @supplierCardinality 0..*
     * @supplierRole types
     */
    private Map types; // Simple name ==> UserTypes

	protected Set importedOnDemandBy; // SourceFiles

	private PackageDecl(PackageDecl owner, String theSimpleName, JstParseTreeBranch theSource) {
		super(owner, theSimpleName, theSource);

        if (owner != null)
	        owner.addSubPackage(this);

		allSources = new HashSet();
        subPackages = new HashMap();
        sourceFiles = new HashSet();
        types = new HashMap();
		importedOnDemandBy = new HashSet();

		addSource(source);
    }

	// Prefix pkg names, because they can coincide with class names.
	public String getCanonicalName() {
    	if (getOwnerDecl() == null || getOwnerDecl() == defaultPackage)
        	return "PKG_" + simpleName;

		return getOwnerDecl().getCanonicalName() + "." + simpleName;
    }

	// Need "normal" name to reflect on classes in imported packages
	public String getQualifiedName() {
    	if (getOwnerDecl() == null)
        	return "";

		String prefix = getOwnerDecl().getOwnerPackage().getQualifiedName();
        if (prefix.length() > 0)
        	prefix += ".";
		return prefix + simpleName;
    }

	// The ownerScope of a package is always null, becauses searches don't propagate up through packages.
	public Scope getOwnerScope() {
		return null;
	}

	public UserType getOwnerUserType() {
		return null;
	}


	public PackageDecl getOwnerPackage() {
		return this;
	}

	public Scope asScope() {
		return this;
	}

	// Generate artificial name for anonymous decl in this scope.
    // Packages do not contain anonymous members.
	public String generateName() {
    	throw new RuntimeException("Package should not contain anonymous declarations");
    }


	/**
     * Create an immediate subpackage.
     * @param source A package statement for the subpackage.
     */
	public PackageDecl addPackage(String name, JstParseTreeBranch source) {
		PackageDecl kid = (PackageDecl) subPackages.get(name);
		if (kid == null)
			kid = new PackageDecl(this, name, source);

		if (source != null)
			kid.addSource(source);

		return kid;
   }

	// Ask a package for a subpackage, but don't create any.
	public PackageDecl lookupPackage(String packageName) {
		return (PackageDecl) subPackages.get(packageName);
    }

	public TypeDecl lookupType(boolean widen, String typeName, ReferenceType looker) {
		return (UserType) types.get(typeName);
    }

	public MethodDecl lookupMethod(boolean widen, String methodName, List params, ReferenceType looker) {
    	return null;
    }

	public VariableDecl lookupVariable(boolean widen, String variableName, ReferenceType looker) {
    	return null;
    }

	// Packages have multiple sources.  Only the first is recorded by the Decl base class.
	protected void addSource(JstParseTreeBranch src) {
    	if (src != null)
	    	allSources.add(src);
    }

    public Set getAllSources() {
        return allSources;
    }

	protected void addSubPackage(PackageDecl kid) {
    	subPackages.put(kid.getSimpleName(), kid);
    }

	public Collection getSubPackages() {
    	return subPackages.values();
    }

	public void addSourceFile(SourceFile sourceFile) {
    	sourceFiles.add(sourceFile);
    }

	public Set getSourceFiles() {
    	return sourceFiles;
    }

	public boolean addType(UserType type) {
    	types.put(type.getSimpleName(), type);
    	return true;
    }

    public Set getTypes() {
        return new HashSet(types.values());
    }

    public void accept(ModelVisitor visitor) {
        visitor.visitPackageDecl(this);
    }
    
	public void addImportedOnDemandBy(SourceFile source) {
		importedOnDemandBy.add(source);
	}

	public Set getImportedOnDemandBy() {
		return importedOnDemandBy;
	}
}
