
package jst.symtab;

import java.lang.reflect.Modifier;
import java.util.*;
import jst.parsetree.JstParseTreeBranch;
import jst.symtab.link.*;

// The ownerScope of a UserType may be a Package (when no source code),
// a Sourcefile, a UserType, or a block.
public abstract class UserType extends ReferenceType implements ImportableScope {

	private static int numAnonymousTypes = 0;
	
	protected int modifier; // A UserType may be a true member.
	
    /**
     *@link aggregation
     *      @associates <{FieldDecl}>
     * @supplierCardinality 0..*
     * @supplierRole fields
     */
    protected Map<String, FieldDecl> fields; // String ==> FieldDecl

    /**
     *@link aggregation
     *      @associates <{MethodDecl}>
     * @supplierCardinality 0..*
     * @supplierRole methods
     */
    protected Map<String, Set<MethodDecl>> methods;

    /**
     *@link aggregation
     *      @associates jst.symtab.UserType
     * @supplierCardinality 0..*
     * @supplierRole types
     */
    protected Map<String, UserType> types; // String ==> UserType

	protected Set importedOnDemandBy; // SourceFiles
	protected Set typeImportedBy; // SourceFiles

	protected UserType(SourceFile ownerSourceFile, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerSourceFile, theSimpleName, theSource);

		ownerSourceFile.addType(this);
    }

	protected UserType(PackageDecl ownerPackage, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerPackage, theSimpleName, theSource);

		ownerPackage.addType(this);
    }

	protected UserType(UserType ownerUserType, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerUserType, theSimpleName, theSource);

		ownerUserType.addType(this);
    }

	protected UserType(Block ownerBlock, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerBlock, theSimpleName, theSource);

		ownerBlock.addType(this);
    }

	protected UserType(FieldInitializer ownerFieldInitializer, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerFieldInitializer, theSimpleName, theSource);

		ownerFieldInitializer.addType(this);
    }

	// Tricksy constructor for ParameterizedClassType subclass.
	protected UserType(GenericClassType theOwner, String theSimpleName, JstParseTreeBranch theSource) {
    	super(theOwner, theSimpleName, theSource);

		// Subclass adds the Parameterized usertype to the Generic owner.
	}

	protected UserType(SourceFile newOwnerSourceFile, UserType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerSourceFile, that, swaps);

		newOwnerSourceFile.addType(this);
		init(that, swaps);
	}

	protected UserType(PackageDecl newOwnerPackage, UserType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerPackage, that, swaps);

		newOwnerPackage.addType(this);
	}

	protected UserType(UserType newOwnerUserType, UserType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerUserType, that, swaps);

		newOwnerUserType.addType(this);
		init(that, swaps);
	}

	protected UserType(Block newOwnerBlock, UserType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerBlock, that, swaps);

		newOwnerBlock.addType(this);
		init(that, swaps);
	}

	protected UserType(FieldInitializer newOwnerFieldInitializer, UserType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerFieldInitializer, that, swaps);

		newOwnerFieldInitializer.addType(this);
		init(that, swaps);
	}

	// Tricksy constructor for ParameterizedClassType subclass.
	protected UserType(GenericClassType newOwner,  UserType that, Map<TypeParameter, ActualType> swaps) {
    	super(newOwner, that, swaps);

		// Subclass adds the Parameterized usertype to the Generic owner.

		init(that, swaps);
	}

	protected void init(UserType that, Map<TypeParameter, ActualType> swaps) {
		modifier = that.modifier;

		for (FieldDecl thatField : that.fields.values())
			thatField.copy(this, swaps);

		for (MethodDecl thatMethod : that.getMethods())
			thatMethod.copy(this, swaps);

		for (TypeDecl thatType : that.types.values())
			thatType.copy(this, swaps);
	}

	public abstract void copy(SourceFile newOwnerSourceFile, Map<TypeParameter, ActualType> swaps);

	public abstract void copy(PackageDecl newOwnerPackage, Map<TypeParameter, ActualType> swaps);

	public abstract void copy(UserType newOwnerUserType, Map<TypeParameter, ActualType> swaps);

	public abstract void copy(Block newOwnerBlock, Map<TypeParameter, ActualType> swaps);

	protected void init() {
		super.init();

        fields = new HashMap<String, FieldDecl>();
        methods = new HashMap<String, Set<MethodDecl>>();
        types = new HashMap<String, UserType>();
		importedOnDemandBy = new HashSet();
		typeImportedBy = new HashSet();
	}

	public UserType asUserType() {
		return this;
	}

	public ClassType asClass() {
		return null;
	}
	
	public InterfaceType asInterface() {
		return null;
	}

	public UserType getNearestUserType() {
		return this;
     }

	public abstract void addSuperInterfaceRef(InterfaceTypeLookup interfaceLookup);

	public void addImportedOnDemandBy(SourceFile source) {
		importedOnDemandBy.add(source);
	}

	public Set getImportedOnDemandBy() {
		return importedOnDemandBy;
	}
	
	public void addTypeImportedBy(SourceFile source) {
		typeImportedBy.add(source);
	}

	public Set getTypeImportedBy() {
		return typeImportedBy;
	}

 	public String getCanonicalName() {
    	if (getOwnerScope() instanceof SourceFile) {
        	PackageDecl pkg = getOwnerPackage();
            if (pkg == PackageDecl.getDefaultPackage())
        		return "TYP_" + simpleName;
            else
       			return "TYP_" + pkg.getCanonicalName().substring(4) + "." + simpleName;
        }


       	return "TYP_" + getOwnerScope().getCanonicalName().substring(4) + "." + simpleName;
    }

	// Generate artificial name for anonymous decl in this scope.
	public String generateName() {
        return "@" + String.valueOf(++numAnonymousTypes);
    }

	public void addField(FieldDecl field) {
		fields.put(field.getSimpleName(), field);
    }

	public Set getFields() {
    	return new HashSet(fields.values());
    }

	public void addMethod(MethodDecl method) {
    	Set overloadedMethods = (Set) methods.get(method.getSimpleName());
        if (overloadedMethods == null) {
        	overloadedMethods = new HashSet();
			methods.put(method.getSimpleName(), overloadedMethods);
        }

		overloadedMethods.add(method);
    }

	public Set<MethodDecl> getMethods() {
     	Set<MethodDecl> result = new HashSet<MethodDecl>();

        // Flatten the sets.
        Iterator setIter = methods.values().iterator();
        while (setIter.hasNext())
        	result.addAll((Set) setIter.next());

     	return result;
    }

	public boolean addType(UserType innerType) {
    	types.put(innerType.getSimpleName(), innerType);
    	return true;
    }

	public Set getTypes() {
    	return new HashSet(types.values());
    }

	public int getModifier() {
    	return modifier;
    }

	public void setModifier(int theModifier) {
    	modifier = theModifier;
    }

	public MethodDecl lookupMethod(boolean widen, String methodName, List params, ReferenceType looker) {
		// Don't worry about overriding here. filterInaccessible removes any doubled-up signatures.
		Set candidates = findAllMethodsNamed(methodName);

		MethodDecl result = (MethodDecl) chooseOperation(candidates, params, looker);
		if (result != null || !widen)
			return result;

		return getOwnerScope().lookupMethod(true, methodName, params, looker);
    }

	protected Set findAllMethodsNamed(String methodName) {		
		int index = methodName.indexOf("(");
		if(index > 0) {
			methodName = methodName.substring(0, index);
		}
		
    	Set result = new HashSet();

		Set localMethods = (Set) methods.get(methodName);
		if (localMethods != null)
			result.addAll(localMethods);

		result.addAll(findInheritedMethodsNamed(methodName));
        return result;
    }

	protected abstract Set findInheritedMethodsNamed(String methodName);

	// Is class accessible to looker?
	public boolean accessible(ReferenceType looker) {
        return looker == null || Modifier.isPublic(modifier) || getOwnerPackage() == looker.getOwnerPackage();
	}

	protected void filterInaccessible(Set operations, ReferenceType looker) {
    	Iterator opIter = operations.iterator();
        while (opIter.hasNext()) {
        	OperationDecl op = (OperationDecl) opIter.next();
            if (!op.accessible(looker))
            	opIter.remove();
        }

		// If two methods have identical signatures, the "lower" one is accessible.
        Set hidden = new HashSet();
    	opIter = operations.iterator();
        while (opIter.hasNext()) {
        	OperationDecl op = (OperationDecl) opIter.next();

			Iterator opIter2 = operations.iterator();
			while (opIter2.hasNext()) {
				OperationDecl op2 = (OperationDecl) opIter2.next();
				if (op != op2 && op.hides(op2))
					hidden.add(op2);
			}
        }

        operations.removeAll(hidden);
	}
	
	 /** New code **********************************************************/
    
    public MethodDecl getOverridden(String methodName, List params, ReferenceType looker) {
        Set candidates = findInheritedMethodsNamed(methodName);
        return chooseOverriddenOperation(candidates, params, looker);
    }
    
    public MethodDecl chooseOverriddenOperation(Set candidates, List params, ReferenceType looker) {
        candidates = new HashSet(candidates); // Avoid changing parameter.
        filterInaccessible(candidates, looker);
        filterInapplicable(candidates, params);
        return (MethodDecl) selectMostSpecific(candidates);
    }
    
    /** New code ends *****************************************************/

	protected void filterInapplicable(Set operations, List params) {
    	Iterator opIter = operations.iterator();
        while (opIter.hasNext()) {
        	OperationDecl op = (OperationDecl) opIter.next();
            if (!op.applicable(params))
            	opIter.remove();
        }
    }

	protected OperationDecl selectMostSpecific(Set operations) {
        Set hidden = new HashSet();
    	Iterator opIter = operations.iterator();
        while (opIter.hasNext()) {
        	OperationDecl op = (OperationDecl) opIter.next();
			Iterator opIter2 = operations.iterator();
			while (opIter2.hasNext()) {
				OperationDecl op2 = (OperationDecl) opIter2.next();
				if (op != op2 && op.isMoreSpecific(op2))
					hidden.add(op2);
			}
        }

        operations.removeAll(hidden);

		if (operations.size() == 1)
        	return (OperationDecl) operations.iterator().next();

		return null;
    }

	public boolean assignableFrom(TypeDecl rhs) {
		if (rhs instanceof NullType)
        	return true;

		if (!(rhs instanceof ReferenceType))
        	return false;

		// Object is a special case: interfaces can be assigned to Object, but don't inherit from it.
        if (rhs.getCanonicalName().equals("TYP_java.lang.Object"))
        	return true;

		ReferenceType refRhs = (ReferenceType) rhs;
        return refRhs.isSubtypeOf(this);
    }

	protected OperationDecl chooseOperation(Set candidates, List params, ReferenceType looker) {
    	candidates = new HashSet(candidates); // Avoid changing parameter.
    	filterInaccessible(candidates, looker);
    	filterInapplicable(candidates, params);
        return selectMostSpecific(candidates);
    }
}
