package parser.parsetree;

import parser.grammar.*;
import java.util.*;
import parser.parsetree.visitor.*;

/**
* SyntaxTreeBranch
* <p>
* There are two main ways to implement parse trees: (1) use a generic node for all
* productions, or (2) use production-specific branches.  The second requires code generation
* from the grammar. E.g. a code generator can subclass ParseTreeBranch to make specific
* parse tree branches, so for decl ::= type variable ;
*     class DeclNode extends ParseTreeBranch {
*     }
*     class DeclNode1 extends DeclNode {
*         TypeNode type;
*         VariableNode variable;
*         ParseTreeLeaf semi;
*     }
*/
    public abstract class AbstractSyntaxTreeBranch implements SyntaxTreeBranch {

    /**
     * Accept a visitor.  GoF Visitor pattern variant.
     *
     * @param visitor the visitor.
     */
    public void accept(SyntaxTreeVisitor visitor) {
        visitor.visitBranch(this);
    }

    /** Getter for SymbolType. */
    public Symbol getSymbolType() {
        return getNonterminalType();
    }

    /**
     * Getter for line on which value starts.
     * <p>
     * Empty trees return 0, because they don't know where they are.
     */
	public int getLine() {
         SyntaxTree kid = getKids()[0];
        return kid == null? 0 : kid.getLine();
    }

   /**
     * Getter for column on which value starts.
     * <p>
     * Empty trees return 0, because they don't know where they are.
     */
	public int getColumn() {
        SyntaxTree kid = getKids()[0];
        return kid == null? 0 : kid.getColumn();
    }

    /** Get the string value of the tree.  I.e. the contents of the leaves. */
    public String getValue() {
        String result = "";
        SyntaxTree[] kids = getKids();

        for (int i = 0; i < kids.length; i++) {
            result += kids[i].getValue();
            if (i+1 < kids.length)
                result += " ";
        }

        return result;
    }

    public SyntaxTreeBranch asBranch() {
        return this;
    }

    public SyntaxTreeLeaf asLeaf() {
        return null;
    }
}
