package parser.pda;

import parser.parsetree.*;
import util.*;
import parser.grammar.*;
import parser.parsetree.treeFactory.*;

import java.util.*;
import parser.pda.tokenStream.Looker;
import parser.pda.stateMachine.StateTable;
import parser.pda.stateMachine.NondeterministicStateTable;

public abstract class RekersParser extends TableDrivenParser {

    protected NondeterministicStateTable stateTable;

    // Map state to "parser" i.e. a top node.
    /** @associates parser.pda.RekersStackNode
     * @supplierRole activeParsers
     * @supplierCardinality 0..**/
	protected Map activeParsers; // Integer ==> Node.

	Token currentToken;
    Set forActor; // Nodes to be reduced.
    Set forShifter; // Nodes to be shifted.

    Set deadEnds; // Nodes.  Used for error reporting.

    public RekersParser(NondeterministicStateTable theStateTable, SyntaxTreeFactory theBuilder) {
	    super(theBuilder);
        stateTable = theStateTable;
    }

    protected StateTable getStateTable() {
        return stateTable;
    }

	// Called at the start of each parse.
	protected void start() {
		activeParsers = new HashMap();
        RekersStackNode node = new RekersStackNode(stateTable.getStartState());
        activeParsers.put(new Integer(stateTable.getStartState()), node);
	}

	protected boolean step(Looker looker) {
        Token token = looker.first();
		currentToken = token;

		forActor = new HashSet(activeParsers.values());
		forShifter = new HashSet();
		deadEnds = new HashSet();

		while (!forActor.isEmpty()) {
			Iterator it = forActor.iterator();
			RekersStackNode p = (RekersStackNode) it.next();
			it.remove();
	
			actor(p);
		}

		// Check for errors before we lose activeParsers info.
		if (forShifter.isEmpty())
			return false;
	
		shifter();
		return true;
    }
	
    protected boolean accept() {
        return activeParsers.get(new Integer(stateTable.getStopState())) != null;
    }

	protected void actor(RekersStackNode p) {
		Symbol tokenType = currentToken.getSymbolType();

		boolean didSomething = false;
	
		if (stateTable.getTransition(p.state, tokenType) != 0) {
			forShifter.add(p);
			didSomething = true;
		}
	
		Production[] reds = stateTable.getReductions(p.state, tokenType);
        if (reds != null)
            for (short i = 0; i < reds.length; i++) {
                Production red = reds[i];
                doReductions(p, red);
                didSomething = true;
            }
	
		if (!didSomething)
			deadEnds.add(p);
    }

	protected void doReductions(RekersStackNode p, Production red) {
		int redSize = red.getSize();
		List kids = new ArrayList(redSize);
        doReductions(p, red, redSize, kids);
    }

    // Rekers algorithm says:
    //   forall p' for which a path of length(red) from p to p' exists do
    //       kids := the tree nodes of the links which form the path from p to p'
    //       REDUCER(p', GOTO(state(p'), ntID), red, kids)
    //
    // In order to find all the paths & populate the kids array for each, I use
    // a recursive algorithm that walks the graph, keeping kids up to date.
	protected void doReductions(RekersStackNode pPrime, Production red, int redSize, List kids) {

		if (redSize == 0) { // We have reached the end of the path.
			Symbol gotoSymbol = red.getLhs();
			int gotoDst = stateTable.getTransition(pPrime.state, gotoSymbol);
			reducer(pPrime, gotoDst, red, kids);
			return;
		}
	
		// Need to back up, collecting kids.
		Iterator prevIter = pPrime.getPrevious().entrySet().iterator();
		while (prevIter.hasNext()) {
			Map.Entry entry = (Map.Entry) prevIter.next();
			RekersStackNode prev = (RekersStackNode) entry.getKey();
			SyntaxTree A = (SyntaxTree) entry.getValue();

			kids.set(redSize - 1, A);
	
			doReductions(prev, red, redSize - 1, kids);
		}
    }

    // REDUCER(p-, state, A::= ff, kids) :
	protected void reducer(RekersStackNode pPrev, int state, Production red, List kids) {
	
		// rulenode := GETRULENODE(A::= ff, kids)
		SyntaxTree rulenode = getRulenode(red, kids);
	
		// if exists p E activeparsers with state(p) = state then
		RekersStackNode p = (RekersStackNode) activeParsers.get(new Integer(state));
		if (p != null) {
			// if there already exists a direct link link from p to p- then
			if (p.hasLink(pPrev))
			// ADDRULENODE(treenode(link), rulenode)
			    addRulenode(p, pPrev, rulenode);
			else {
				// n := GETSYMBOLNODE(A, rulenode)
				// XXX SyntaxTreeNode n = getSymbolnode(red, kids);
		
				// add a link link from p to p- with tree node n
				// XXX p.addLink(pPrev, n);
				p.addLink(pPrev, rulenode);

				// Get the top nodes that we're not already planning on reducing
				//
				// forall p' in (activeparsers - foractor) do
				Set pPrimes = new HashSet(activeParsers.values());
				pPrimes.removeAll(forActor);
		
				Iterator pPrimeIter = pPrimes.iterator();
				while (pPrimeIter.hasNext()) {
					RekersStackNode pPrime = (RekersStackNode) pPrimeIter.next();
					// Get the reductions for pPrime.
					// forall (reduce rule) E ACTION(state(p'), currenttoken) do 
					Production[] reds = stateTable.getReductions(pPrime.state, currentToken.getSymbolType());
					for (short i = 0; reds != null && i < reds.length; i++) {
						Production rule = reds[i];
						// DOLIMITEDREDUCTIONS(p', rule, link)
						doLimitedReductions(pPrime, rule, p, pPrev);
					}
				}
			}
		}
		else {
			// create a stack node p with state state
			// n := GETSYMBOLNODE(A, rulenode)
			// add a link from p to p- with tree node n
	
			// XXX SyntaxTreeNode n = getSymbolnode(red, kids);
			// XXX p = new Node(pPrev, n, state);
			p = new RekersStackNode(pPrev, rulenode, state);

			// add p to activeparsers
			activeParsers.put(new Integer(state), p);
	
			// add p to foractor
			forActor.add(p);
		}
    }

    // Cut'n'paste unlimited version, with added params.
	protected void doLimitedReductions(RekersStackNode p, Production red, RekersStackNode linkFrom, RekersStackNode linkTo) {
		int redSize = red.getSize();
		List kids = new ArrayList(redSize);
        doLimitedReductions(p, red, redSize, kids, linkFrom, linkTo);
    }

    // Cut'n'paste unlimited version, with added params.
    // The link is represented by from & to nodes, which go to null when that link is found.
    //
    // DOLIMITEDREDUCTIONS(p, A::= ff, link) :
    // forall p' for which a path of length(ff) from p to p' through link exists do
    //     kids := the tree nodes of the links which form the path from p to p'
    //		REDUCER(p', GOTO(state(p'), A), A::= ff, kids)

	protected void doLimitedReductions(RekersStackNode pPrime, Production red, int redSize, List kids,
	    RekersStackNode linkFrom, RekersStackNode linkTo) {

		if (redSize == 0) { // We have reached the end of the path.
	
			if (linkFrom != null) // We didn't hit the required link.
			    return;
	
			Symbol gotoSymbol = red.getLhs();
			int gotoDst = stateTable.getTransition(pPrime.state, gotoSymbol);
			reducer(pPrime, gotoDst, red, kids);
			return;
		}
	
		// Need to back up, collecting kids.
		Iterator prevIter = pPrime.getPrevious().entrySet().iterator();
		while (prevIter.hasNext()) {
			Map.Entry entry = (Map.Entry) prevIter.next();
			RekersStackNode prev = (RekersStackNode) entry.getKey();
			SyntaxTree A = (SyntaxTree) entry.getValue();
	
			if (pPrime == linkFrom && prev == linkTo) { // This is the required link.
				linkFrom = null; // No longer required.
				linkTo = null;
			}
	
			kids.set(redSize - 1, A);
	
			doLimitedReductions(prev, red, (short) (redSize - 1), kids, linkFrom, linkTo);
		}
    }
		

    protected void shifter() {
		// activeparsers := 0
		activeParsers.clear();
	
		// create a term node n with token currenttoken
		Token n = currentToken;
	
		// forall <p-, state'> E forshifter do
		Iterator prevIter = forShifter.iterator();
		while (prevIter.hasNext()) {
			RekersStackNode pPrev = (RekersStackNode) prevIter.next();
			int statePrime = stateTable.getTransition(pPrev.state, currentToken.getSymbolType());
	
			// if exists p E activeparsers with state(p) = state' then
			RekersStackNode p = (RekersStackNode) activeParsers.get(new Integer(statePrime));
			if (p != null)
			// add a link from p to p- with tree node n
			p.addLink(pPrev, n);
			else {
			// create a stack node p with state state'
			// add a link from p to p- with tree node n 
			p = new RekersStackNode(pPrev, n, statePrime);

			// add p to activeparsers
			activeParsers.put(new Integer(statePrime), p);
			}
		}
    }
		
    // GETRULENODE(r, kids) :
	protected SyntaxTree getRulenode(Production red, List kids) {
		// return a rule node with rule r and elements kids
		return treeBuilder.makeBranch(red, kids);
	}

	    
    // ADDRULENODE(symbolnode, rulenode) : 
	protected void addRulenode(RekersStackNode linkFrom, RekersStackNode linkTo, SyntaxTree rulenode) {
		// add rulenode to the possibilities of symbolnode
        AbstractSyntaxTreeBranch existing = (AbstractSyntaxTreeBranch) linkFrom.getPrevious().get(linkTo);
        //linkFrom.previous.put(linkTo, builder.makeAmbiguity(existing, (SyntaxTreeBranch) rulenode));
        treeBuilder.makeAmbiguity(existing, (AbstractSyntaxTreeBranch) rulenode);
    }
	    
    // GETSYMBOLNODE(s, rulenode) :
	protected SyntaxTree getSymbolnode(Production red, List kids) {
		// return a symbol node with symbol s and possibilities { rulenode }     
		return getRulenode(red, kids);
    }

    protected SyntaxTreeBranch getResult() {
        RekersStackNode eofNode = (RekersStackNode) activeParsers.get(new Integer(stateTable.getStopState()));
        if (eofNode == null)
            return null;

        Set prev = eofNode.getPrevious().keySet();
        assert prev.size() == 1;
        RekersStackNode rootNode = (RekersStackNode) prev.iterator().next();

        Collection trees = rootNode.getPrevious().values();
        assert trees.size() == 1;

        SyntaxTree tree = (SyntaxTree) trees.iterator().next();
        return tree.asBranch();
    }

    protected Set getExpected() {
        Set result = new HashSet();

        Iterator deadIter = deadEnds.iterator();
	    while (deadIter.hasNext()) {
	        RekersStackNode top = (RekersStackNode) deadIter.next();

            result.addAll(stateTable.getExpected(top.getState()));
        }

        return result;
	}

    protected String getContext() {
        String result = "";

        Iterator deadIter = deadEnds.iterator();
	    while (deadIter.hasNext()) {
	        RekersStackNode top = (RekersStackNode) deadIter.next();

            result += stateTable.getContext(top.getState()) + " ";
        }

        return result;
	}
}
